-- YOU WILL NEED TO PLACE YOUR PROGRAM TO THE ROM STORE ARRAY
-- BUT SHOULD NOT ALTER ANYTHING ELSE IN THIS FILE
-- UNLESS YOU ARE VERY SURE OF WHAT YOU ARE DOING
use work.cpu_package.all;
entity rom is
   port(rom_sel: in bit;
        address_in: in address;
        rom_ready: out bit;
        instr: inout bus_word bus);
end rom;

architecture behaviour of rom is
begin
  rom: process
    variable store: program_array:=
    -- INSERT YOUR PROGRAM HERE, LIKE THIS:
        (0 => B"0100_0000_0000_0001", -- ld r0, X"0100" (file address)
         1 => X"0100",
         2 => B"0100_0001_0000_0001", -- ld r1, X"0F00" (bottom of RAM)
         3 => X"0F00",
         4 => B"0001_0000_0001_0010", -- add r0,r1,r2
         5 => B"0011_0010_0001_0000", -- st  r2,(r1)
         6 => B"0011_0010_0000_0000", -- st  r2,(r0) (save memory to disk)
         others => B"0000000000000000"); -- rest nops
    -- LEAVE EVERYTHING BELOW THIS UNTOUCHED 
    variable current_address: address_val; 
    variable store_loc: natural; 
    begin
       wait until rom_sel = '1';
       current_address := address_to_nat(address_in);
       store_loc := natural(current_address);
       instr <= store(store_loc);
       rom_ready <= '1' after 5 ns;
       wait until rom_sel = '0';
       instr <= X"0000"; -- stop driving the bus now
       rom_ready <= '0' after 5 ns;
end process;
end behaviour;


use work.cpu_package.all;
entity decoder is
       port(io_req: in bit; -- from CPU
            address_in: in address;  -- from CPU
            rom_ready, ram_ready: in bit; -- from devices
            ready: out bit; -- to CPU
            rom_sel, ram_sel, file_sel: out bit -- to devices
            );
end decoder;

architecture behaviour of decoder is
begin
  decode: process
   variable current_address: address_val;
   begin
      wait until io_req = '1';
      current_address := address_to_nat(address_in);
     -- check address is a valid one
      assert (current_address <= (rom_top+1))or(current_address>=ram_bottom)
               report "Invalid address selected";

     -- if address ok, select correct device
      if current_address <= rom_top then   -- must be ROM address
         rom_sel <= '1';
         wait until rom_ready = '1';
         -- now handle handshaking back to CPU;
         ready <= '1';
         wait until io_req = '0';
         rom_sel <= '0';
         ready <= '0' after 5 ns;
      elsif current_address >= ram_bottom then -- must be RAM address
         ram_sel <= '1';
         wait until ram_ready = '1';
         -- now handle handshaking back to CPU;
         ready <= '1';
         wait until io_req = '0';
         ram_sel <= '0';
         ready <= '0' after 5 ns;
      elsif current_address = rom_top+1 then -- file write is just above ROM
         file_sel <= '1';
         wait until ram_ready = '1';
         -- now handle handshaking back to CPU;
         ready <= '1';
         wait until io_req = '0';
         file_sel <= '0';
         ready <= '0' after 5 ns;
      end if;
   end process;
end behaviour;



use work.cpu_package.all;
use std.textio.all;
entity ram is
   port(ram_sel, file_sel, r_w: in bit;
        address_in: in address;
        ram_ready: out bit;
        mem_data: inout bus_word bus);
end ram;

architecture behaviour of ram is

    -- procedure to save ram contents to disk
    procedure file_save(store: ram_array) is
       file ram_file: text is out "ramfile";
       variable int_data: integer;
       variable bv_data: word;
       variable addr: natural:=0;
       variable next_line: line;
       variable space: string(1 to 2) := ": ";
       begin
            for i in 0 to (ram_size-1) loop
                bv_data := store(addr);  -- retrieve a word
                write(next_line,addr);
                write(next_line,space);
                write(next_line,bv_data);
                writeline(ram_file,next_line);
                addr := addr+1;
            end loop;
    end file_save;

begin
  memory: process 
    variable store: ram_array;
    variable current_address: address_val; 
    begin
       wait until (ram_sel = '1') or (file_sel = '1');
       if ram_sel = '1' then
           current_address := address_to_nat(address_in)-ram_bottom;
           if r_w = '1' then -- must be read
              mem_data <= store(natural(current_address));
              ram_ready <= '1' after 5 ns;
              wait until ram_sel = '0';
              mem_data <= X"0000"; -- stop driving the bus
              ram_ready <= '0' after 5 ns;
           else -- must be write
              store(natural(current_address)) := mem_data;
              ram_ready <= '1' after 5 ns;
              wait until ram_sel = '0';
              ram_ready <= '0' after 5 ns;
           end if;
       elsif file_sel = '1' then  -- dump memory to disk
             file_save(store);
             ram_ready <= '1' after 5 ns;
             wait until file_sel = '0';
             ram_ready <= '0' after 5 ns;
      end if;
end process;
end behaviour;  