/*
 *    vhd.c is taken from pas.c and modified to handle VHDL language
 *    contructs.
 *
 *    7/27/94 Rewritten by Carlos Vedovato, Silicon Arts
 *
 *    Original source code from Premia:
 *
 *                Copyright 1991,1992,1993 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN

#define MAX_BUF_LEN    100

/* all 'VHDL' keyword names. */
static LPVOID _vhd_hash_names        = NIL;  /* pointer to the hash table containing */
static LPVOID _vhd_user_names_1        = NIL;  /* pointer to the hash table user keyword names. */
static LPVOID _vhd_user_names_2        = NIL;  /* pointer to the hash table user keyword names. */

/* an abbreviated name and its template */
static LPVOID _vhd_hash_templates    = NIL;  /* pointer to the hash table containing */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _vhd_hash_templates.
 */
typedef struct _LanguageTemplate {         
 LPSTR abbrev;
 LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR VHD_INDENT_STR = "\377BEGIN\377";

/*
 * _vhd_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _vhd_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 *     %x - where x is a number 0-9, inserts a macro defined in cwright.ini
 *          e.g. %0=' ', %1='\t', %2='@'
 */
LanguageTemplate _vhd_templates[] = {
 { "if",        "if & then\n%1\n@end if;" },
 { "be|beg|begi|begin",    "begin\n%1&\n@end;" },
 { "for",       "for & in  loop\n%1\n@end loop;" },
 { "while",       "while & loop\n%1\n@end loop;"},
 { "ifgen",       "if & generate\n%1\n@end generate;" },
 { "forgen",       "for & in  generate\n%1\n@end generate;" },
 { "pro|proc|proce|proces|process", "process(&)\nbegin\n%1\n@end process;" },
 { "proced|procedu|procedur|procedure","procedure &(\n%1\n@) is\nbegin\n%1\n@end;" },
 { "conf|config|configuration",  "configuration & of  is\n%1\n@begin\n%1\n@end;" },
 { "ent|enti|entit|entity",   "entity & is\n%1port (\n%1\n@);\n@end ;" },
 { "com|comp|component",    "component &\n%1port(\n%1\n@);\n@end component;" },
 { "arc|arch|architecture",   "architecture & of  is\n%1\n@begin\n%1\n@end ;" },
 { "ca|cas|case",     "case & is\n%1when  => ;\n@end case;" },
 { "const|constant",     "constant & : " },
 { "arr|array",      "array & of " },
 { "sig|signal",      "signal & : " },
 { "var|variable",     "variable & : " },
 { "waitf",       "wait for & ns;"},
 { "libi",       "Library IEEE;\nuse IEEE.std_logic_1164.all;\n"},
 { "std",       "STD_LOGIC;"},
 { "stdv",       "STD_LOGIC_VECTOR(&);"},
 { NIL,       NIL },
};


/* Define the list of 'VHDL' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _vhd_names[] = {
 "abs",       "access",     "after",       "alias",      "all",
    "and",       "architecture", "array",     "assert",     "attribute",
    "begin",     "block",      "body",        "buffer",     "bus",
    "case",      "component",  "configuration", "constant", "disconnect",
    "downto",    "else",       "elsif",       "end",        "entity",
    "exit",      "file",       "for",         "function",   "generate",
    "generic",   "guarded",    "if",          "in",         "inout",
    "is",        "label",      "library",     "linkage",    "loop",
    "map",       "mod",        "nand",        "new",        "next",
    "nor",       "not",        "null",        "of",         "on",
    "open",      "or",         "others",      "out",        "package",
    "port",      "procedure",  "process",     "range",      "record",
    "register",  "rem",        "report",      "return",     "select",
 "severity",  "signal",     "subtype",     "then",       "to",
 "transport", "type",       "units",       "until",      "use",
 "variable",  "wait",       "when",        "while",      "with",
 "xnor",      "xor",

 NIL,
};


static long _vhd_in_comment( long offset );

void DLL _vhd_init( void );
BOOL DLL _vhd_expand( void );
long DLL _vhd_language_colors_update( long firstline, long lastline );
void DLL _vhd_language_colors( long numlines );
void DLL _vhd_assign_template( LPSTR abbrev, LPSTR template );
void DLL _vhd_add_keyword( LPSTR keyword, int table );
BOOL DLL _vhd_indent( void );


void DLL 
_init()
{
    _vhd_init();
}

/*
 ** _vhd_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_vhd_init( void )
{
 static BOOL firstTime = TRUE;
 int    i;

 if (firstTime)
 {
  LibExport( "BOOL _vhd_expand");
  LibExport( "_vhd_assign_template LPSTR LPSTR");
  LibExport( "_vhd_add_keyword LPSTR int");
  LibExport( "BOOL _vhd_indent( void ) ");
  LibExport( "long _vhd_language_colors_update long long");
  LibExport( "_vhd_language_colors long");
  LibExport( "_vhd_routines");

  firstTime = FALSE;

  /* Create a hash table to hold all of the 'VHDL' keyword names.
   */
  _vhd_hash_names = HashCreateTable( 41 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
  i = 0;

  while ( _vhd_names[ i ] != NIL )
  {
   LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _vhd_hash_names, 
          _vhd_names[i], strlen( _vhd_names[i] ) );
   if (lpf)
    *lpf = StrNew( _vhd_names[i] );
   i++;
  }

  /* Create a hash table to hold all of the 'VHDL' template names.
   */
  _vhd_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), 
HASH_IGN_CASE );
  i = 0;

  while ( _vhd_templates[ i ].abbrev != NIL )
  {
   _vhd_assign_template( _vhd_templates[i].abbrev, _vhd_templates[i].template );
   i++;
  }
 }
}


/*
 ** _vhd_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'VHDL' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_vhd_expand( void )
{
 LPVOID htEntry;
 LPSTR  _vhd_string = NULL;
 LPSTR  str;
 int    i = 0;

 MarkSavePos();

 /* Check to see if we are currently positioned inside a 'VHDL' comment
  */
 if (_vhd_in_comment( -1L ) != -1)
  goto _vhd_expand_incomplete;

 str = BufReadStr( 65535 );
 while (str[i] && !isspace(str[i]))
  i++;
 str[i] = '\0';

 _vhd_string = StrTrim( str, " \t" );
 StrFree( str );

 if (!_vhd_string)
  goto _vhd_expand_incomplete;

 /* Search for a keyword match in the hash table.
  */
 htEntry = HashFindEntry( _vhd_hash_templates, _vhd_string, strlen( _vhd_string ) );

 if (htEntry)
 {
  /* A template exists for the current keyword, delete the keyword
   * and insert the template.
   */
  MarkDropPos( 1 );
  str = *(LPSTR*)htEntry;
  BufDelChar( lstrlen( _vhd_string ) );

  _ext_expand_template( str );
 } 
 else
  goto _vhd_expand_incomplete;
 StrFree( _vhd_string );
 return TRUE;

_vhd_expand_incomplete:
 StrFree( _vhd_string );
 MarkRestorePos();

 return FALSE;
}

/*
 ** _vhd_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_vhd_language_colors_update( long firstline, long lastline )
{
 long offset,offset_to_restore;
 long count;

 if (firstline && lastline)
 {
  offset_to_restore = BufQOffset();
  _PosInit( 0 );

  if (firstline > 1)  {
   _PosNextLine( firstline - 1);
  }
  offset = _PosQOffset();

  AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );

  _PosInit( offset );
  _vhd_language_colors( lastline - firstline + 1 );

  MovToOffset( offset_to_restore );
 }
 count = (lastline - firstline);
 return(count);
}

/*
 ** _vhd_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_vhd_language_colors( long numlines )
{
 long  startoff;
 long  startline;
 long  endoff;
 long  inc;
 WORD  ch;
 LPSTR buf;
 int   i = 0;
 int   comment_color = ColorComments( -1 );
 int   keyword_color = ColorKeywords( -1 );
 int   quote_color      = ColorAlternate1( -1 );
 int   number_color     = ColorAlternate2( -1 );
 int   color;

 buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

 /* Start at the top of the buffer or current position 
  */
 if (numlines == -1)
 {
  _PosInit( 0 );  
  numlines = 0x7FFFFFFF;
 }
 else
 {
  startoff  = _PosQOffset();
  startline = _PosQLine();
  endoff    = _vhd_in_comment( startoff );
  if (endoff == -1)
   _PosInit( startoff );
  else
  {
   _PosInit( endoff );
   numlines += (startline - _PosQLine());
  }
 }
 ch = _PosCurrentChar();
 startoff = _PosQOffset();


 while ((ch != EOF_CHAR) && (numlines > 0))
 {
  if (ch == '\n')
   numlines--;

  if (ch == '"')
  {
   UINT quote = ch;

     startoff = _PosQOffset();

   do 
   {
    ch = _PosNextChar();
    if (ch == '\\')
    {
     ch = _PosNextChar();
     if ((ch == quote) || (ch == '\\'))
      ch = _PosNextChar();
    }
    if ((ch == EOF_CHAR) || (ch == '\n'))
     break;
   } while (ch != quote);

   if (ch == quote)
    ch = _PosNextChar();

   /* color the string */
   if (quote_color)
   {
    endoff = _PosQOffset();
   _PosInit( startoff );
    _PosSetColor( quote_color, endoff - startoff );
    _PosInit( endoff );
   }
   continue;
  }
  else if (ch == '\'') 
  {
      startoff = _PosQOffset();
      endoff = -1;
      ch = _PosNextChar();
      if (ch == '\\')
      {
      	ch = _PosNextChar();
      	ch = _PosNextChar();
      	if (ch == '\'')
      	{
      		ch = _PosNextChar();
      		endoff = _PosQOffset();
      	}
      	else
      	{
      		ch = _PosPrevChar();
      		ch = _PosPrevChar();
      	}
      }
      else if (ch != EOF_CHAR)
      {
      	ch = _PosNextChar();
      	if (ch == '\'')
      	{
      		ch = _PosNextChar();
      		endoff = _PosQOffset();
      	}
      	else
      	{
      		ch = _PosPrevChar();
      	}
      }
      if (endoff > startoff)
      {
      	_PosInit( startoff );
      	_PosSetColor( quote_color, endoff - startoff );
      	_PosInit( endoff );
      }
      continue;
  }
  else if (number_color && isdigit( ch ))
  {
     int j = 0;
   UINT lastch = _PosPrevChar();

   if (lastch != EOF_CHAR)
    _PosNextChar();

   if ((  lastch != ' '
    && lastch != '\t'
    && lastch != '\r'
    && lastch != '\n'
    && lastch != EOF_CHAR
    && !ispunct( lastch ))
    || lastch == '_') {
     ch = _PosNextChar();
     continue;
    }

   startoff = _PosQOffset();
   do
   {
    j++;
    ch = _PosNextChar();
   } while (isxdigit( ch ) || (ch == '.') || (ch == 'x' && j == 1));

   if (ch == '_') {
     continue;
   }

   if (ch == 'l' || ch == 'L')
    ch = _PosNextChar();

   endoff = _PosQOffset();
   _PosInit( startoff );
   _PosSetColor( number_color, endoff - startoff );
   _PosInit( endoff );
   continue;
  }
  else if (isalnum( (BYTE)ch ) || (ch == '_'))
  {
   WORD lastch = _PosPrevChar();
 
   if (lastch != EOF_CHAR)
    _PosNextChar();
 
   i = 0;
   startoff = _PosQOffset();
 
   do
   {
    if (i < MAX_BUF_LEN)
     buf[i++] = (char)ch;
    ch = _PosNextChar();
   } while (isalpha( (BYTE)ch ) || (ch == '_'));
 
   if ((lastch != ' ' 
      && lastch != '\t' 
      && lastch != '\r' 
      && lastch != '\n' 
      && lastch != EOF_CHAR
      && !ispunct( lastch ))
     || (ch == '_'))
    i = 0;
   continue;
  }
  else if (i)
  {
//   if (HashFindEntry( _vhd_hash_names, buf, i ))
//   {
//    endoff = _PosQOffset();
//    _PosInit( startoff );
//    _PosSetColor( keyword_color, endoff - startoff );
//    _PosInit( endoff );
//   }
//   i = 0;

     STRUPR(buf);
     if (_vhd_user_names_1 && HashFindEntry( _vhd_user_names_1, buf, i ))
       	color = ColorAlternate5( -1 );
     else if (_vhd_user_names_2 && HashFindEntry( _vhd_user_names_2, buf, i ))
      	color = ColorAlternate6( -1 );
     else if (HashFindEntry( _vhd_hash_names, buf, i ))
     	color = keyword_color;
     else
     	color = 0;

     if (color)
     {
     	endoff = _PosQOffset();
       	_PosInit( startoff );
       	_PosSetColor( color, endoff - startoff );
       	_PosInit( endoff );
     }
     i = 0;
   
  } 

  if (ch == '-')
  {
   startoff = _PosQOffset();
   ch = _PosNextChar();

   if (ch == '-')
   {
    inc = _PosNextLine( 1 );
    numlines--;
    endoff = _PosQOffset();
    _PosInit( startoff );
    _PosSetColor( comment_color, endoff - startoff - inc );
    _PosInit( endoff );
    ch = _PosCurrentChar();
   }
   continue;
  }

  ch = _PosNextChar();
 }
 MemFree( buf );
}

void DLL
_vhd_add_keyword( LPSTR keyword, int table )
{
// if (keyword && *keyword)
// {
//  LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _vhd_hash_names, 
//        keyword, strlen( keyword ) );
//  if (lpf)
//   *lpf = StrNew( keyword );
// }
	if (keyword && *keyword)
	{
		LPSTR XFAR *lpf;
		LPVOID hash_table;

		switch( table )
		{
		case 0:
			hash_table = _vhd_hash_names;
			break;
		case 1:
			if (!_vhd_user_names_1)
				_vhd_user_names_1 = HashCreateTable( 41, /* a prime # */
						sizeof( LPSTR ), 0 );
			hash_table = _vhd_user_names_1;
			break;
		case 2:
			if (!_vhd_user_names_2)
				_vhd_user_names_2 = HashCreateTable( 41, /* a prime # */
						sizeof( LPSTR ), 0 );
			hash_table = _vhd_user_names_2;
			break;
		default:
			hash_table = NIL;
			break;
		}
		if (hash_table)
		{
			lpf = (LPSTR XFAR *)HashGetEntry( hash_table, 
					STRUPR(keyword), STRLEN( keyword ) );
			if (lpf)
				*lpf = StrNew( keyword );
		}
	}

}


/*
 ** _vhd_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'VHDL' template strings
 *    without changing the code.
 */
void DLL
_vhd_assign_template( LPSTR abbrev, LPSTR template )
{
 _ext_assign_template( _vhd_hash_templates, abbrev, template );
}


/*
 ** _vhd_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'VHDL' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in VHD_INDENT_STR - perform smart
 *       indentation.  This may include skipping the next line if
 *       it consists only of an 'BEGIN' statement.
 */
BOOL DLL
_vhd_indent( void )
{
 long   curline;
 long   col;
 LPSTR  _vhd_string;
 LPSTR  str;
 long   matchpos;

 MarkSavePos();

 curline = BufQCurrentLine();

 /* Check to see if we are currently positioned inside a 'VHDL' comment
  */
 if (_vhd_in_comment( -1L ) != -1)
  goto _vhd_indent_incomplete;

 /* Not in a comment so indent according to the current 'VHDL' construct.
  */
 if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
   || (BufQCurrentLine() != curline) )
  goto _vhd_indent_incomplete;

 /* Read the keyword at the beginning of the current line.
  */
 col = BufQCurrentCol();
 SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
 str = BufReadStr( col - BufQCurrentCol() );
 _vhd_string = StrTrim( str, " \t" );
 StrFree( str );

 /* Set prefix and suffix to \377 to perform a quick lookup of the name
  */
 str = StrNew( "\377" );
 _vhd_string = StrAppend( _vhd_string, str );
 str = StrAppend( str, _vhd_string );
 StrFree( _vhd_string );
 
 matchpos = StrMatch( str, VHD_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );

 StrFree( str );

 if (!matchpos)
  /* Not a 'VHDL' keyword, do normal indentation. */
  goto _vhd_indent_incomplete;
  
 MarkRestorePos();

 _ext_insert_indented_EOL();
 BufInsertChar( '\t' );
 return TRUE;

 _vhd_indent_incomplete:
  MarkRestorePos();

 return FALSE;
}



/*--  Private functions --*/


/*
 ** _vhd_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'VHDL' comment by search backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_vhd_in_comment( long offset )
{
 WORD  ch;

 /* Check to see if we are currently positioned inside a 'VHDL' comment
  */
 _PosInit( offset );
 ch = _PosCurrentCharNL();

 do
 {
  switch( ch )
  {
  case '-':
   ch = _PosPrevCharNL();
   if (ch == '-')    /* Found the beginning of a comment */
    return _PosQOffset();  /* currrent line is in a comment */
   break;
  default:
   ch = _PosPrevCharNL();
   break;
  }
 } while ((ch != EOF_CHAR) && (ch != EOL1_CHAR) && (ch != EOL2_CHAR));

 return -1L;
}


void DLL
_vhd_routines( void )
{
 if (Grep("^(([Pp](ROCEDURE|rocedure))|([Ff](unction|UNCTION)))[ \t]+[a-zA-Z0-9_]+.*$"))
  MsgMessage("");
 else
  MsgMessage("No routines found.");
}
